
  R%=3
REM   PROCtest(1024)
  mem%=1024*64
  PROCmeasure(mem%)
END

DEF PROCfast_sieve(a()%, l%)
  PRINT "l%=";l%
  CALL fast_sieve%, a()%, l%
  PRINT "   -> l%="; l%
ENDPROC

REM In BBC Basic: DIM a%(N%) declares "int a[N+1]", understood as in C
REM               i.e. the indexes a[0] to a[N] are valid (inclusively)
REM               n%=DIM(a%(),1) return n% set to N%


DEF PROCmeasure(mem%)
  LOCAL quantum%, i%, t0%, t1%, t2%, dt%, f%, msg$, slow, fast, rep%, l%, old_l%, r%, dt_slow, dt_fast
  DIM a%((mem%+3)/4-1)
  PRINT "   ======== Sieve of Eratosthenes Benchmark v1.8 (BBC Basic variant) ========"
  PRINT "Benchmarking in a ";mem%/1024;" KB block of memory"

  quantum%=1<<30
  FOR i%=0 TO 5
    t0%=TIME
    REPEAT
      t1%=TIME
    UNTIL t0%<>t1%
    dt%=t1%-t0%
    IF quantum%>dt% THEN
      quantum%=dt%
    ENDIF
  NEXT i%    
  PRINT "Timer resolution is ";10000*quantum%;" us"
  
  REM f%=OPENOUT "eratosth/txt"
  f%=OPENOUT "RAM::RamDisc0.eratosth/txt"  
  BPUT#f%,"# Interpreted by BBC BASIC V"
  msg$="# Benchmarking on a "+STR$(mem%/1024)+" KB block with timer resolution of "+STR$(10000*quantum%)+" us"
  BPUT#f%,msg$
  BPUT#f%,"# slow_version is using global variables, fast_version only local variables"
  BPUT#f%,"#    size  slow_version [us] fast_version [us]  # of inner iterations"
                    
  l%=30
  WHILE l%<=mem%
    slow=1E20
    fast=1E20
    rep%=0
  
    FOR i%=1 TO R%
      t0%=TIME
      IF i%=1 THEN
        WHILE TIME-t0%<256*quantum%
          PROCglobal_sieve(a%(),l%)
          rep%+=1
        ENDWHILE      
      ELSE
        FOR r%=1 TO rep%
          PROCglobal_sieve(a%(),l%)
        NEXT r%
      ENDIF
      t1%=TIME
      FOR r%=1 TO rep%
        PROCsieve(a%(),l%)
      NEXT r%
      t2%=TIME
      dt_slow=(t1%-t0%+0.0)/rep%
      dt_fast=(t2%-t1%+0.0)/rep%
      IF slow>dt_slow
        slow=dt_slow
      ENDIF
      IF fast>dt_fast
        fast=dt_fast
      ENDIF
    NEXT i%
    
    PRINT "Progress: ";l%;" of ";mem%;" = "; (l%*100.0)/mem%  ;"%, rep="; rep%
    msg$=STR$(l%)+" "+STR$(10000*slow)+" "+STR$(10000*fast)+" "+STR$(FNct_sieve(a%(),l%))
    BPUT#f%,msg$
        
    old_l%=l%
    l%=(11*l%)/10
    IF l%<=old_l% THEN 
      l%=l%+1
    ENDIF
  ENDWHILE

  CLOSE#f%
END


DEF PROCprint_primes(f%,a%())
  LOCAL n% : LOCAL N% : LOCAL k% : LOCAL index% : LOCAL mask%
  n%=1+DIM(a%(),1)
  N%=32*n%-1
  FOR k%=0 TO N%
    index% = k%>>5
    mask% = (1<<(k% AND 31))
    IF (a%(index%) AND mask%) = 0 THEN
      BPUT#f%, STR$(k%)
      PRINT k%;" ";
    ENDIF
  NEXT k%
  PRINT
END

DEF PROCtest(n%)
  DIM a%((n%+3)/4-1)
  PROCsieve(a%(), 4*(1+DIM(a%(),1)))
  CALL fast_sieve% a%() n%
  f%=OPENOUT "primes"
  PROCprint_primes(f%,a%())
  CLOSE#f%
END

DEF FNct_sieve(a%(), l%)
  LOCAL n% : LOCAL N% : LOCAL k% : LOCAL ct% : LOCAL index% : LOCAL mask%
  n%=(l% + 3) >> 2  
  N%=32*n%
  a%() =&55555555
  a%(0)=&55555553
  k%=3
  ct%=0
  WHILE k%*k%<N%
    index% = k%>>5
    mask% = (1<<(k% AND 31))
    IF (a%(index%) AND mask%) = 0 THEN
      FOR i% = 2*k% TO N%-1 STEP k%
        index% = i%>>5
        mask% = (1<<(i% AND 31))
        a%(index%) = a%(index%) OR mask%
        ct%+=1
      NEXT i%
    ENDIF
    k%+=1
  ENDWHILE
  =ct%

DEF PROCsieve(a%(), l%)
  LOCAL n% : LOCAL N% : LOCAL k% : LOCAL i% : LOCAL index% : LOCAL mask%
  n%=(l% + 3) >> 2  
  N%=32*n%
  a%() =&55555555
  a%(0)=&55555553
  k%=3
  WHILE k%*k%<N%
    index% = k%>>5
    mask% = (1<<(k% AND 31))
    IF (a%(index%) AND mask%) = 0 THEN
      FOR i% = 2*k% TO N%-1 STEP k%
        index% = i%>>5
        mask% = (1<<(i% AND 31))
        a%(index%) = a%(index%) OR mask%
      NEXT i%
    ENDIF
    k%+=1
  ENDWHILE
ENDPROC


DEF PROCglobal_sieve(a%(), l%)
  n%=(l% + 3) >> 2  
  N%=32*n%
  a%() =&55555555
  a%(0)=&55555553
  k%=3
  WHILE k%*k%<N%
    index% = k%>>5
    mask% = (1<<(k% AND 31))
    IF (a%(index%) AND mask%) = 0 THEN
      FOR i% = 2*k% TO N%-1 STEP k%
        index% = i%>>5
        mask% = (1<<(i% AND 31))
        a%(index%) = a%(index%) OR mask%
      NEXT i%
    ENDIF
    k%+=1
  ENDWHILE
ENDPROC


