
; AMD64 implementation

	bits 64
	
	global fast_sieve

	section	.text
	align 32

;****** x86-64 Linux calling convetions (the subset needed here) ******
;
; callee-saved registers:               rbx rbp rsp r12 r13 r14 r15
; argument registers (in order):        rdi rsi rdx rcx r8  r9
; temporary registers:                  rax r10 r11
;
; return value registers:               rax rdx
;

; void fast_sieve(u8* a, ux n)
;                 rdi    rsi

	
fast_sieve:

	push	rbx
	push	rbp

	mov	rcx, rsi	; rsi=n, rcx=n/2; 
	shr	rcx, 1
	lea	r10, [rsi*8]	; N = 8*n
	mov	rdx, rdi	; rdi = a
	add	rdx, rsi	; rdx = b = a+n

	mov	rbp, rdi	; rbp = a
	mov	eax, 0x5555
	cld
	rep	stosw
	test	esi, 1
	jz	even
	stosb
even:
	mov	BYTE [rbp], 0x53	; a[0]=0x53
	mov	rdi, rbp		; rdi = a
	mov	rbp, rdx		; rbp = b


outer_loop:
	mov	esi, 2			; rsi = k = 2
	lea	r11, [2*rsi]		; r11 = kk = 4

continue:
	lea	rcx, [1+2*rsi]		; rcx=1+2*k
	inc	rsi			; k++
	add	r11, rcx		; kk+=rcx
	cmp	r11, r10
	jae	quit

	mov	rdx, rsi
	mov	eax, 1
	mov	ecx, esi
	sar	rdx, 3			; rdx=k>>3
	and	ecx, 7			; al=1<<(k&7)
	sal	eax, cl
	test	BYTE [rdi+rdx], al
	jnz	continue
	
	lea	rbx, [2*rsi]
	sub	rbx, r10		; rbx=i=2*k-N
	jns	continue
	align 32
inner_loop:		
    ; INPUT rbx=i rsi=k rbp=b
    ; USING rax=1<<(i&7) rbx=i rcx=i&7 rdx=i>>3 rsi=k rbp=b 
    ; FREE  rdi r10 r11
%rep 0				; loop unrolling disabled as it leads to no speedup
	mov	rdx, rbx	; on Phenom-IIX4-980 and likely on most modern CPUs
	mov	eax, 1
	mov	ecx, ebx
	sar	rdx, 3			; rdx=i>>3
	and	ecx, 7
	sal	eax, cl
	or	BYTE [rbp+rdx], al	; rdx=i>>3
	add	rbx, rsi		; rbx=i, rsi=k
	jns	continue
%endrep
	mov	rdx, rbx
	mov	eax, 1
	mov	ecx, ebx
	sar	rdx, 3			; rdx=i>>3
	and	ecx, 7
	sal	eax, cl
	or	BYTE [rbp+rdx], al	; rdx=i>>3
	add	rbx, rsi		; rbx=i, rsi=k
	js	inner_loop
	jmp 	continue

quit:
	pop	rbp
	pop	rbx
	ret


