// >>>>> sieve.cc


/* Completely stand-alone reimplementation for my ARM-based development board.
 * 
 * axrun -l sieve-ARM7TDMI.cc | tee ARM7TDMI-60MHz
 *
 */

#define Hz 60000000
#define CLOCKS_PER_TICK 2
#include <at91sam7s.cc>
#include <bios.cc>
#include <printf.cc>
#include <startup.cc>
#define printf dprintf
typedef float F;
inline F log(F _){return _;} /* dummy */
inline void free(void* a){}  /* dummy */

FIQ("	ldr	pc, [pc,#-0xF20]") 
"	ldr     r9, _last_time
	mvn     r8, #"IMM((~TC_CV(OS_TC_TIMER))&0xFFFFF000)"
	ldr     r12, [r8,#(-0xFFF+"IMM(TC_CV(OS_TC_TIMER)&0xFFF)")]
	sub     r12, r12, r9
	mov     r12, r12, lsl #16
	add     r9, r9, r12, lsr #16
	str     r9, _last_time
	add	r9, r9, #0xE000
	str     r9,[r8,#(-0xFFF+"IMM(TC_RC(OS_TC_TIMER)&0xFFF)")]
	ldr     r12,[r8,#(-0xFFF+"IMM(TC_SR(OS_TC_TIMER)&0xFFF)")]
	subs	pc, r14, #4
"
FEND

TT  last_time SECTION(".fiq")=0;

void HW_set_timer(TT timeout, TT last)
{
  out(TC_RC(OS_TC_TIMER), last+timeout);
  in(TC_SR(OS_TC_TIMER));
  out(AIC_ICCR, TC(OS_TC_TIMER));
}

void armboard_init() SECTION(".fiq");
void armboard_init()
{
  out(AIC_IDCR, u32(-1));
  enter_CPU_mode(LOCKED_MODE|DISABLE_IRQ|DISABLE_FIQ);
  HW_set_timer(0xA000, FRD(last_time));
  init_os_timer_and_irqs(u32(-1));
  start_me_up(12,20*3,0);
  UNLOCK;
}

inline TT atmel_HW_time() {IFT(atmel_HW_time)
  TT r;
  asm volatile("
        mvn     %0, #"IMM((~TC_CV(OS_TC_TIMER))&0xFFFFF000)"
        ldr     %0, [%0,#(-0xFFF+"IMM(TC_CV(OS_TC_TIMER)&0xFFF)")]
  " : "=r" (r) : );
  return r;
}

inline TT HW_time(TT last)
{IFT(HW_time)
  TT hw_now=atmel_HW_time();
  UNLOCK;
  return last + u16(hw_now - last);
}

TT time() SECTION(".fiq");
TT time()
{
  LOCK;
  return HW_time(FRD(last_time));  //UNLOCK je uvnitr HW_time()
}

inline int TT2us(TT tt)
{
  return 100*tt/(Hz/(10000*CLOCKS_PER_TICK));
}



#include "sieve.h"

#ifdef ASSEMBLY

void fast_sieve(u8* a, ux n)    /* a[] must be 32bit aligned! */
{
  asm volatile("
	stmfd	sp!, {r4, r5, r6, r7, r8, r10, lr}
	mov	r10, r0
	cmp	r1, #0
	ldmlefd sp!, {r4, r5, r6, r7, r8, r10, pc}

	add	r7, r0, r1
	sub	r7, r7, #(4*7)
	mov	r2, #0x55
	orr	r2, r2, r2, lsl #8
	orr	r2, r2, r2, lsl #16
	mov	r3, r2
	mov	r4, r2
	mov	r5, r2
	mov	r6, r2
	mov	r8, r2
	mov	r12, r2
fastfill:
	stmia	r10!, {r2,r3,r4,r5,r6,r8,r12}
	cmp	r10, r7
	ble	fastfill
	add	r7, r7, #(4*7)		@ now r7 points after the array
	cmp	r10, r7
	bge	filled
slowfill:
	strb	r2, [r10], #1
	cmp	r10, r7
	blt	slowfill
filled:
	mov	r5, #1			@ always 1
	mov	r4, r1, asl #3		@ r4=N=8*n
	add	r8, r0, r1		@ r8=b=a+n 
	mov	r1, #9			@ r1=kk=3*3, n is not accesible from now on
	cmp	r1, r4
	mov	r2, #0x53
	strb	r2, [r0]
	ldmcsfd	sp!, {r4, r5, r6, r7, r8, r10, pc}

	mov	r10, #3			@ r10=k=3
loop:
	and	r2, r10, #7
	ldrb	r3, [r0, r10, lsr #3]
	ands	r3, r3, r5, asl r2
	bne	continue
	rsbs	r12, r4, r10, asl #1	@ r12=(2*k)-N
	bpl	continue
inner_loop:
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bpl	continue
	and	r7, r12, #7
	ldrb	r2, [r8,+r12, asr #3]
	orr	r2, r2, r5, asl r7
	strb	r2, [r8,+r12, asr #3]
	adds	r12, r12, r10
	bmi	inner_loop
continue:
	add	r1, r1, r10, asl #1	@ kk+=2*k+1   (r1=kk)
	add	r1, r1, #1
	add	r10, r10, #1		@ k++         (r10=k)
	cmp	r1, r4
	bcc	loop
	ldmfd	sp!, {r4, r5, r6, r7, r8, r10, pc}");
}

#endif

void measure()
{
  const u32 n=56000;
  u32 a32[(n+3)/4]; u8* a=(u8*)a32;  /* to enforce 32bit alignment */
  TT slow, fast;
  printf("# ARM7TDMI at AT91SAM7S256 (64 KB RAM) running at %d Hz\n", Hz);
  printf("#    size  slow_version [us] fast_version [us]  # of inner iterations\n");
  for(u32 l=30; l<=n; l=max(l+1,11*l/10)){
//  for(u32 l=30; l<=n; l=max(l+1,31*l/30)){
    slow=fast=(u32(1)<<31)-1;
    for(int i=2; i>=0; i--){
      TT t0=time();
      sieve(a,l);
      TT t1=time();
      fast_sieve(a,l);
      TT t2=time();
      slow=min(slow, t1-t0);
      fast=min(fast, t2-t1);
    }
    printf("%9u %17d %17d %l17d  TT=%d\n", l, TT2us(slow), TT2us(fast), ct_sieve(a,l), time());
  }
}

#if 1   /* 1 to benchmark, 0 to print the prime numbers */

int main()
{
  armboard_init();
  measure();
}

#else

void print_primes(u8* a, ux n)
{
  ux N=8*n;
  for(ux i=2, ct=1; i<N; i++) if(!(a[i>>3]&(1<<(i&7)))){
    printf("%d %d\n", u32(i), u32(ct));
    ct++;
  }
}

int main()
{
  const ux n=56000;
  u32 a32[(n+3)/4]; u8* a=(u8*)a32;  /* to enforce 32bit alignment */
  for(int k=0; k<=4*7*2; k++){
    fill(a,u8(0xFF),n);
    fast_sieve(a, n-k);
    if(k>0) printf("---- k=%d\n",k);
    print_primes(a, n-k);
  }
  return 0;
}

#endif

