// >>>>> sieve.cc

#define VERSION "v1.11"

/*

Compile:
 g++ -O3 -g0 -fomit-frame-pointer --unsigned-char -fexpensive-optimizations -o sieve sieve.cc

Run:
 sieve > D

Plot:
 gnuplot
 p "D" using (8*$1):($2/$4) with lines, "D" using (8*$1):($3/$4) with lines

or use plot or plot2 scripts in the RESULTS directory.

 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#ifdef AMIGA           // for gcc-2.95 on amiga we pretend we are Unix
  #undef AMIGA
  #define __unix__
#endif

#if defined(__STORM__) // Amiga StormC 3.0s  //!!! doplnit High-C
  #define AMIGA
#endif

#if defined(__unix__) || defined(AMIGA) ///!!!! doplnit macos X
  #define __PREEMPTIVE_MULTITASKING_OS__
#endif

#ifdef __PREEMPTIVE_MULTITASKING_OS__
 #define R 5  /* How many times to repeat a measurement. The best time is */
#else         /* taken from those trials. Has to be higher in multi tasking */
 #define R 2  /* systems to hide interference from other tasks, and must be */
#endif        /* at least 2. */

#ifdef __BORLANDC__
  #define __16bit_SYSTEM__
#endif

#define STRINGIFY_expand(nr) #nr
#define STRINGIFY(nr) STRINGIFY_expand(nr)
typedef double F;
typedef unsigned char ch;
typedef unsigned char u8;
typedef signed char i8;
typedef short int i16;
typedef unsigned short int u16;

#ifdef __16bit_SYSTEM__
  typedef long int i32;   /* have to use long int to obtain i32 on 16bit system */
  typedef unsigned long int u32;
  #define PD "ld"         /* and the long version of %d formatting modifiers */
  #define PU "lu"
#else
  typedef int i32;        /* this is standard these days */
  typedef unsigned int u32;
  #define PD "d"
  #define PU "u"
#endif

#ifdef __GNUC__
  typedef long long int i64;
  typedef unsigned long long int u64;
  #define PLLU "llu"
#else
  typedef i32 i64;  /* for compilers that don't support 64 bit integers, just fake them */
  typedef u32 u64;
  #define PLLU PU
#endif

#ifdef __GNUC__
 #if (ULONG_MAX) == 0xFFFFFFFFFFFFFFFFUL
   #define __64bit_SYSTEM__   /* otherwise we assume a 32 bit system */
 #elif (ULONG_MAX) != 0xFFFFFFFFUL
   #error "sizeof(unsigned long int) is neither 4 or 8"
 #endif
#endif

#ifndef __64bit_SYSTEM__
 typedef signed long int ix;         /* 32 bit */
 typedef unsigned long int ux;
#else
 typedef signed long long int ix;    /* 64 bit */
 typedef unsigned long long int ux;
#endif
#if defined(Windows_CE)
  #undef max	/* under WinCE we got to get rid of these macros first */
  #undef min
#endif
inline ux max(ux a, ux b) {return a<b?b:a;}
inline ux min(ux tx, ux ty) {return tx<ty ? tx : ty;}
inline F fmin(F tx, F ty) {return tx<ty ? tx : ty;}

#ifdef __HIGHC__

  #include <timeb.h>
  static inline u64 utime0()
  {
    struct timeb tv;
    ftime(&tv);
    return u32(tv.time)*1000000 + 1000*u32(tv.millitm);
  }

  static u64 last=0;

  /* The MetaWare C timer under MS DOS has very poor resolution of 50 ms
     (sometimes only 60 ms) and it even jumps to the past from time to time!
     That's why we try to correct it by only accepting a value that is
     higher than the previously read one.
   */
  static inline u64 utime1()
  {
    u64 t;
    do{
      t=utime0();
    }while(i32(t-last)<0 && last-t<(1<<30));
    last=t;
    return t;
  }

  /* Last try to find a more precise timer in DOS. Unfortunately, it turned
     out that this one also has a resolution of 50 ms. At least it does not
     jump backwards.
   */
  extern "C" {u32 clock(void);}
  static inline u64 utime()
  {
    return 10000*u64(clock());
  }

#elif defined(AMIGA) // Amiga StormC 3.0s

  #include <devices/timer.h>
  #include <proto/timer.h>
  #include <clib/timer_protos.h>
  #include <clib/dos_protos.h>
  inline i64 utime()
  {
    struct timeval tv;
    GetSysTime(&tv);
    return u64(tv.tv_secs)*1000000 + tv.tv_micro;
  }

#elif defined(__BORLANDC__)

  #include <dos.h>
  #include <conio.h>

  static inline i64 utime()
  {
    struct time tv;
    gettime(&tv);
    return tv.ti_hund*u64(10000) + (u64(tv.ti_sec) + (tv.ti_min+tv.ti_hour*u64(60))*60)*u64(1000000);
  }

#elif defined(__unix__) || defined(__HP_aCC)  /* Unix OS (only GNU/Linux and Irix tested so far) or HPUX */

  #include <sys/time.h>
  #include <unistd.h>
  static inline i64 utime()
  {
    struct timeval tv;
    gettimeofday(&tv,0);
    return tv.tv_usec + (i64(tv.tv_sec)*1000000);
  }
#elif defined(Windows_CE)

  #include <Winbase.h>
  #define utime utime1	 /* to avoid clash with windows utime symbol */
  inline i64 utime()
  {
    return u64(GetTickCount())*1000;
  }

#else

  #error "This platform is not supported yet."

#endif

inline void fill(u8 *dest, u8 val, ix len)
{ memset(dest, val, len);}

void arr(u8* a, ux n)
{
  for(ux i=0; i<n; i++) printf("%x\n", a[i]);
}

void print_primes(u8* a, ux n)
{
  ux N=8*n;
  for(ux i=2, ct=1; i<N; i++) if(!(a[i>>3]&(1<<(i&7)))){
//    printf("%d %d %g %g\n", u32(i), u32(ct), i/log(F(i)), (i/log(F(i)))/ct);
//    printf("%d %d\n",u32(i), u32(ct));
    printf("%"PLLU"\n",u64(i));
    ct++;
  }
}


#include "sieve.h"

void measure(ux n)
{
  u8* a;
  for(;;){
    a=(u8*)malloc(n);
    if(a) break;
    n=u32((u64(n)*10)/11);
  }
  fill(a,u8(0),n); /* to map the memory now so that OS would not interfere
		    * with the measurement later
		    */
  fprintf(stderr,"   ======== Sieve of Eratosthenes Benchmark "VERSION" ========\n"
		 "Benchmarking in a %"PD" KB block of memory (\"./sieve N\" changes this to N)\n", n/1024);
  u32 quantum=u32(1)<<31;
  for(int i=5; i>=0; i--){
    u32 t0=utime(), t1;
    do{
      t1=utime();
    }while(t1==t0);
    quantum=min(quantum,t1-t0);
  }
  fprintf(stderr,"Timer resolution is %"PU" us\n", quantum);
#ifdef __unix__
  fprintf(stderr,"DON'T FORGET TO RUN \"full_power\" SCRIPT AS ROOT BEFORE TESTING ON LINUX!\n");  fflush(stderr);
#endif
  printf(
         "# Sieve of Eratosthenes Benchmark "VERSION"\n"
#ifdef ASSEMBLY
         "# Hand-tuned assembly implementation of the fast_version.\n"
#endif
#ifdef __GNUC__
         "# Compiled by gcc-"STRINGIFY(__GNUC__)"."STRINGIFY(__GNUC_MINOR__) "." STRINGIFY(__GNUC_PATCHLEVEL__) ".\n"
#elif defined(__HIGHC__)
         "# Compiled by Metaware High C R3.03\n"
#elif defined(__BORLANDC__)
         "# Compiled by Borland C++ 3.1\n"
#endif
	 "# Benchmarking on a %"PU" KB block with timer resolution of %"PU" us.\n", n/1024, quantum);
  printf("#    size  slow_version [us] fast_version [us]  # of inner iterations\n");
#if defined(__PREEMPTIVE_MULTITASKING_OS__) && !defined(AMIGA)
  sleep(1);    /* to let the OS do its chores after it mapped the memory */
#endif
  for(ux l=30; l<=n; l=max(l+1,11*l/10)){
    F slow, fast;
    slow=fast=1e20;
    u32 rep=0;
    for(int i=R-1; i>=0; i--){
      u64 t0=utime();
      if(i==R-1){
	for(; utime()-t0<256*quantum; rep++) sieve(a,l);
      }else{
	for(i32 r=rep-1; r>=0; r--) sieve(a,l);
      }
      i64 t1=utime();
      for(i32 r=rep-1; r>=0; r--) fast_sieve(a,l);
      i64 t2=utime();
      slow=fmin(slow, F(t1-t0)/rep);
      fast=fmin(fast, F(t2-t1)/rep);
    }
    printf("%9"PLLU" %17.3f %17.3f %17"PLLU"\n", u64(l), slow, fast, ct_sieve(a,l)); fflush(stdout);
    fprintf(stderr, "Progress: %"PLLU" of %"PLLU" = %5.2f%%, rep=%"PU" \r", u64(l), u64(n), F(l)*100/n, rep); fflush(stderr);
#ifdef __BORLANDC__
    if(kbhit()) exit(1); /* to make it interruptible */
#endif
#ifdef AMIGA
    if(CheckSignal(SIGBREAKF_CTRL_C))
    {
      fprintf(stderr, "\nBREAK!\n");
      exit(0);
    }
#endif
  }
  fprintf(stderr,"\nFinished.\n"); fflush(stderr);
  free(a);
}

#ifdef Windows_CE   /* Windows CE support added by Jiri Zima */
int WINAPI WinMain(HINSTANCE hInstance,	HINSTANCE hPrevInstance,
                   LPTSTR lpCmdLine,	int nCmdShow)
{
  u32 n=16*1024*1024;
  measure(n);
  return 0;
}
#else
int main(int argc, char** argv)
{
  char* pn=*argv;
  argv++;
#ifdef __16bit_SYSTEM__
  u32 n=56*u32(1024);
#else
  ux n=256*1024*1024;
#endif
  ux m=n;
  u8 flag=0;
  for(;*argv; argv++){
    if(*argv[0]=='-'){
      if(strcmp(*argv,"-h")==0 || strcmp(*argv,"--help")==0 || strcmp(*argv,"-?")==0){
      help:
        fprintf(stderr,"Sieve of Eratosthenes Benchmark "VERSION"\n"
                       "Usage: %s [MKB] > CPUname-Cfg-FreqMHz   runs the sieve benchmark\n"
                       "    or %s -h | --help | -?              prints this help\n"
                       "    or %s -v                            prints version\n"
                       "    or %s -l                            prints prime numbers\n"
                       "    or %s -a                            prints the array in hex\n"
                       "  where MKB is the maximum size of the array in KiB, i.e. the largest\n"
                       "  number that will be tested is 8*1024*MKB-1.\n"
                       "  Without arguments, %"PLLU" KiB would be tested.\n",
                       pn,pn,pn,pn,pn,u64(m>>10));
        return 0;
      }else if(strcmp(*argv,"-v")==0){
        fprintf(stderr,VERSION"\n");
        return 0;
      }else if(strcmp(*argv,"-l")==0){
        flag|=1;
      }else if(strcmp(*argv,"-a")==0){
        flag|=2;
      }else{
        fprintf(stderr,"unknown option %s\n", *argv);
        return 1;
      }
    }else{
      n=atoll(*argv)*1024;
      if(n==0) goto help;
    }
  }
  if(!flag)
    measure(n);
  else{
    u8* a;
    fprintf(stderr,"Sieve array %"PLLU" bytes long.\n", u64(n)); fflush(stdout);
    a=(u8*)malloc(n);
    fast_sieve(a,n);
    if(flag&2) arr(a,n);
    if(flag&1) print_primes(a,n);
    free(a);
  }  
  return 0;
}
#endif

