
#ifndef FAST_VERSION
 #define FROM 2*k
#else
 #define FROM k*k
#endif

/* Originally I tested from 2*k, which is redundant (testing from k*k works
   just as well). I realized that only after making lots of measurements and
   it even turned out that running the test using k*k blurs the impact of
   L1/L2/L3 caches in the plot. So I decided to keep it from 2*k for
   benchmarking purposes (unless the macro FAST_VERSION is defined in case
   of comparing this code against a different sieve implementation (let's
   say in a different language)).
 */
   

u64 ct_sieve(u8*a, ux n)
{
  ux N=8*n;
  ux k, i;
  fill(a, u8(0x55), n);
  a[0]=0x53;
  u64 ct=0;
  for(k=3; k*k<N; k++){
    if(a[k>>3]&(1<<(k&7))) continue;
    for(i=FROM; i<N; i+=k){
      a[i>>3]|=1<<(i&7);
      ct++;
    }
  }
  return ct;
}

void sieve(u8* a, ux n)
{
  ux N=8*n;
  ux k, i;
  fill(a, u8(0x55), n);
  a[0]=0x53;
  for(k=3; k*k<N; k++){
    if(a[k>>3]&(1<<(k&7))) continue;
    for(i=FROM; i<N; i+=k)
      a[i>>3]|=1<<(i&7);
  }
}


#ifndef ASSEMBLY

/* This is optimized version. Should be faster but can only address half of the
   memory the sieve() function can. Normally this is not a problem except if ux
   and ix types were 16bits. To avoid these problems they are defined as 32bit
   even on 16bit machines.
 */
void fast_sieve(u8* a, ux n)
{
  ux N=8*n;
  u8* b=a+n;
  ux k, kk;
  ix i;
  fill(a, u8(0x55), n);
  a[0]=0x53;
  for(k=3, kk=3*3; kk<N; kk+=2*k+1, k++){
    if(a[k>>3]&(1<<(k&7))) continue;
    for(i=FROM-N; i<0; i+=k)  /* signed i limits range to half of sieve() */
      b[i>>3]|=1<<(i&7);
  }
}

#else

extern "C" void fast_sieve(u8* a, ux n);

#endif

