

function sieve(n::Int)
  isp=ones(Bool,n)
  isp[1]=false;
  for j in 2:2:n isp[j]=false; end
  for i in 3:round(Int,sqrt(n))
    if(isp[i])
      for j in (2*i):i:n
        isp[j]=false
      end
    end
  end
  return isp
end


function fast_sieve(N::Int)
  m=div(N+7,8)
  a=Array(UInt8,m)
  @inbounds @simd for i in 2:m a[i]=0x55; end
  a[1]=0x53;
  k=3;
  ct=0;
  while k*k<N
    @inbounds if((a[1+(k>>3)]&(1<<(k&7)))==0)
      for i in (2*k):k:(N-1)
        @inbounds a[1+(i>>3)]|=1<<(i&7)
      end
    end
    k+=1
  end
  return a;
end

function sieve_ct(N::Int)
  m=div(N+7,8)
  a=Array(UInt8,m)
  for i in 2:m a[i]=0x55; end
  a[1]=0x53;
  k=3;
  ct=0;
  while k*k<N
    if((a[1+(k>>3)]&(1<<(k&7)))==0)
      for i in (2*k):k:(N-1)
        ct+=1
        a[1+(i>>3)]|=1<<(i&7)
      end
    end
    k+=1
  end
  return ct;
end



# show(filter( x->sieve(1000)[x], 1:1000 ));


n=256779580;
l=30
R=10
while l<=n
  dt_slow=1<<30
  dt_fast=1<<30
  sct=sieve_ct(8*l);
  for r in 1:R
    tic()
    sieve(8*l)
    dt_slow=min(dt_slow,toq())
    tic()
    fast_sieve(8*l)
    dt_fast=min(dt_fast,toq())
  end
  print("$l $(dt_slow*1e6) $(dt_fast*1e6) $sct\n");
  l=max(l+1,div(11*l,10))
end

